#!/usr/bin/python

from __future__ import print_function

import pyBigWig as pbw
import negspy.coordinates as nc
import clodius.tiles as ct
import h5py
import math
import numpy as np
import os
import os.path as op
import pandas as pd
import sys
import time
import argparse

def reduce_data(data_array):
    s = set(data_array)
    lookup_table = dict([(x,i) for i,x in enumerate(s)])

    print("len lookup_table:", len(lookup_table), "len data_array:", len(data_array))
    t1 = time.time()
    new_data = [lookup_table[d] for d in data_array]
    t2 = time.time()
    
    print("Set size:", len(s), 'data size:', len(data_array))
    print("time taken:", 1000000 * (t2 - t1) / len(data_array))
    return (np.array(new_data), list(s))


def main():
    parser = argparse.ArgumentParser(description="""
    
    python main.py
""")

    parser.add_argument('filepath')
    parser.add_argument('-c', '--chunk-size', default=14, type=int)
    parser.add_argument('-z', '--zoom-step', default=8, type=int)
    parser.add_argument('-t', '--tile-size', default=1024, type=int)
    parser.add_argument('-o', '--output-file', default=None)
    parser.add_argument('-a', '--assembly', default='hg19')

    args = parser.parse_args()
    last_end = 0
    data = []

    if args.output_file is None:
        args.output_file = op.splitext(args.filepath)[0] + '.hitile'

    print("output file:", args.output_file)

    # Override the output file if it existts
    if op.exists(args.output_file):
        os.remove(args.output_file)
    f = h5py.File(args.output_file, 'w')

    # get the information about the chromosomes in this assembly
    chrom_info = nc.get_chrominfo(args.assembly)
    chrom_order = nc.get_chromorder(args.assembly)
    assembly_size = chrom_info.total_length

    tile_size = args.tile_size
    chunk_size = tile_size * 2**args.chunk_size     # how many values to read in at once while tiling

    dsets = []     # data sets at each zoom level

    # initialize the arrays which will store the values at each stored zoom level
    z = 0
    positions = []   # store where we are at the current dataset
    data_buffers = [[]]

    while assembly_size / 2 ** z > tile_size:
        dsets += [f.create_dataset('values_' + str(z), (assembly_size / 2 ** z,), dtype='f',compression='gzip')]
        data_buffers += [[]]
        positions += [0]
        z += args.zoom_step

    # load the bigWig file
    print("filepath:", args.filepath)
    bwf = pbw.open(args.filepath) 

    # store some meta data
    d = f.create_dataset('meta', (1,), dtype='f')

    d.attrs['zoom-step'] = args.zoom_step
    d.attrs['max-length'] = assembly_size
    d.attrs['assembly'] = args.assembly
    d.attrs['chrom-names'] = bwf.chroms().keys()
    d.attrs['chrom-sizes'] = bwf.chroms().values()
    d.attrs['chrom-order'] = chrom_order
    d.attrs['tile-size'] = tile_size
    d.attrs['max-zoom'] = max_zoom =  math.ceil(math.log(d.attrs['max-length'] / tile_size) / math.log(2))
    d.attrs['max-width'] = tile_size * 2 ** max_zoom

    print("assembly size (max-length)", d.attrs['max-length'])
    print("max-width", d.attrs['max-width'])
    print("max_zoom:", d.attrs['max-zoom'])
    print("chunk-size:", chunk_size)
    print("chrom-order", d.attrs['chrom-order'])

    t1 = time.time()

    for chrom in nc.get_chromorder(args.assembly):
        if chrom not in bwf.chroms():
            print("skipping chrom (not in bigWig file):", chrom)
            continue

        counter = 0
        chrom_size = bwf.chroms()[chrom]

        while counter < chrom_size:
            remaining = min(chunk_size, chrom_size - counter)
            values = bwf.values(chrom, counter, counter + remaining)
            #print("counter:", counter, "remaining:", remaining, "counter + remaining:", counter + remaining)
            #print("values:", values)

            counter += remaining
            curr_zoom = 0
            data_buffers[0] += values
            curr_time = time.time() - t1
            percent_progress = (positions[curr_zoom] + 1) / float(assembly_size)
            print("progress: {:.2f} elapsed: {:.2f} remaining: {:.2f}".format(percent_progress,
                curr_time, curr_time / (percent_progress) - curr_time))

            while len(data_buffers[curr_zoom]) >= chunk_size:
                # get the current chunk and store it, converting nans to 0
                curr_chunk = np.array(data_buffers[curr_zoom][:chunk_size])
                curr_chunk[np.isnan(curr_chunk)] = 0
                dsets[curr_zoom][positions[curr_zoom]:positions[curr_zoom]+chunk_size] = curr_chunk

                # aggregate and store aggregated values in the next zoom_level's data
                data_buffers[curr_zoom+1] += list(ct.aggregate(curr_chunk, 2 ** args.zoom_step))
                data_buffers[curr_zoom] = data_buffers[curr_zoom][chunk_size:]
                positions[curr_zoom] += chunk_size
                data = data_buffers[curr_zoom+1]
                curr_zoom += 1

    # store the remaining data
    print("tile_size:", tile_size, positions[0])

    while True:
        # get the current chunk and store it
        chunk_size = len(data_buffers[curr_zoom])
        curr_chunk = np.array(data_buffers[curr_zoom][:chunk_size])
        dsets[curr_zoom][positions[curr_zoom]:positions[curr_zoom]+chunk_size] = curr_chunk

        print("curr_zoom:", curr_zoom, "position:", positions[curr_zoom] + len(curr_chunk))
        print("len:", [len(d) for d in data_buffers])

        # aggregate and store aggregated values in the next zoom_level's data
        data_buffers[curr_zoom+1] += list(ct.aggregate(curr_chunk, 2 ** args.zoom_step))
        data_buffers[curr_zoom] = data_buffers[curr_zoom][chunk_size:]
        positions[curr_zoom] += chunk_size
        data = data_buffers[curr_zoom+1]
        curr_zoom += 1

        # we've created enough tile levels to cover the entire maximum width
        if curr_zoom * args.zoom_step >= max_zoom:
            break

    # still need to take care of the last chunk

    data = np.array(data)
    t1 = time.time()

    '''
    curr_zoom = 0
    dsets = []
    while len(data) > tile_size:
        (to_store_data, values_list) = reduce_data(data)
        if len(values_list) < 2 **16 and len(data) > 2**20:
            print "storing lookup"
            dsets += [f.create_dataset("zoom_" + str(curr_zoom), (len(data),), dtype='i2', compression='gzip')]
            lookup_dset = f.create_dataset("values_" + str(curr_zoom), (len(values_list), ), dtype='f')
            lookup_dset[:] = np.array(values_list)

            dsets[curr_zoom][:len(data)] = to_store_data
        else:
            print "storing raw...."
            dsets += [f.create_dataset(str(curr_zoom), (len(data),), dtype='f4', compression='gzip')]
            dsets[curr_zoom][:len(data)] = data

        new_data = ct.aggregate(data, 32)
        data = new_data
        curr_zoom += 1

    dsets += [f.create_dataset(str(curr_zoom), (len(data),), dtype='i2', compression='gzip')]
    dsets[curr_zoom][:len(data)] = data

    print "time:", time.time() - t1
    print "len(data):", len(data)
    '''

    #print data

if __name__ == '__main__':
    main()





import math
import numpy as np
cimport numpy as np

def aggregate(np.ndarray[np.float32_t,ndim=1] in_array, int num_to_agg):
    '''
    Calculate a new array which contains the sums of every
    num_to_agg elements of the original array.

    Example:

    aggregate([1,2,3,4], 2) #=> [3,7]

    :param in_array: A numpy array
    :param num_to_agg: The number of elements to aggregate into one
    :return: A numpy array
    '''
    cdef int length = len(in_array)
    cdef np.ndarray[np.float32_t,ndim=1] out_array = np.zeros(int(math.ceil(length / float(num_to_agg))), 
                                                              dtype=np.float32)
    cdef int i = 0
    cdef int j = 0

    #print "length:", length, "num_to_agg:", num_to_agg, "new length:", len(out_array), math.ceil(length / num_to_agg) , 

    while i < length:
        #print "i/num_to_agg:", length, i, i / num_to_agg
        out_array[i / num_to_agg] = 0;
        j = 0
        while j < num_to_agg and (i+j) < length:
            out_array[i / num_to_agg] += in_array[i+j]
            j += 1
        i += num_to_agg

    return out_array



from __future__ import print_function

import json
import math
import numpy as np
import sys
import time
import clodius.fast as cf

def aggregate(in_array, num_to_agg):
    return cf.aggregate(in_array.astype(np.float32), num_to_agg)

def load_entries_from_file(sc, filename, column_names=None, delimiter=None, 
        elasticsearch_path=None):
    '''
    Load a dataset from file.

    :parma sc: SparkContext (or FakeSparkContext)
    :param filename: The filename for the file (either JSON or tsv) containing
        the data
    :param column_names: If the passed file is a tsv, then we may want to pass
        in column_names. If column_names is None then we assume that the file 
        contains column_names. column_names should be an array.
    :return: An array of dictionaries containing the data.
    '''

    def add_column_names(x):
        return dict(zip(column_names, x))

    if delimiter is not None:
        entries = sc.textFile(filename).map(lambda x: dict(zip(column_names, x.strip().split(delimiter))))
    else:
        entries = sc.textFile(filename).map(lambda x: dict(zip(column_names, x.strip().split())))
    
    return entries

def expand_range(x, from_col, to_col, range_except_0 = None):
    '''
    Copy a row multiple times if two columns indicate that it
    represents a range of values. The most common example would 
    be a bed file.

    chr1 1000 2000 4.6

    Means that the nucleotides between 1000 and 2000 on chromosome 1
    have a value of 4.6
    '''
    new_xs = []
    if range_except_0 is not None:
        if x[range_except_0] == '0':
            return []

    for i in range(int(x[from_col]), int(x[to_col])):
        new_x = x.copy()
        new_x[from_col] = i
        new_x[to_col] = i+1
        new_xs += [new_x]
    return new_xs

def summarize_data(max_entries):
    '''
    Summarize the data into a maximum of max_entries.

    :return: A function that can be called on a dataset to 
    condense it into a maximum of max_entries entries.
    '''
    def condense(data):
        random.shuffle(data)
        return data[:max_entries]

    return condense

def data_bounds(entries, num_dims):
    '''
    Get the minimum and maximum values for a data set.

    :param entries: A list of dictionaries representing the data
    :param num_dims: The number of dimensions in the data set
    :return: (mins, maxs)
    '''

    mins = [entries.map(lambda x: x['pos'][i]).reduce(reduce_min) for i in range(num_dims)]
    maxs = [entries.map(lambda x: x['pos'][i]).reduce(reduce_max) for i in range(num_dims)]

    return (mins, maxs)

def add_pos(dim_names, add_uuid=False):
    def add_pos_func(entry):
        new_dict = entry
        new_dict['pos'] = [float(entry[dn]) for dn in dim_names]

        if add_uuid:
            new_dict['uuid'] = shortuuid.uuid()

    return add_pos_func

def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z

def make_tiles_by_importance(sc,entries, dim_names, max_zoom, mins, maxs, importance_field=None,
        max_entries_per_tile=10, output_dir=None, gzip_output=False, add_uuid=False,
        reverse_importance=False, end_dim_names=None, adapt_zoom=True):
    '''
    Create a set of tiles by restricting the maximum number of entries that
    can be shown on each tile. If there are too many entries that are assigned
    to a particular tile, the lower importance ones will be removed.

    If no importance is given, the entries will be selected randomly.

    :param sc: SparkContext
    :param entries: A list of dictionary containing the data which we wish to tile
    :param dim_names: The names of the dimensions along which we wish to break
                      up the tiles
    :param max_zoom: If it's None, then we will automatically calculate the maximum
                     zoom level such that all entries are shown.
    :param importance_field: The field which contains the importance of the entries.
    :return: A set of tiles
    '''
    if end_dim_names is None:
        end_dim_names = dim_names


    entries = entries.map(lambda x: merge_two_dicts(x, {'pos': [float(x[dn]) for dn in dim_names]}))
    entries = entries.map(lambda x: merge_two_dicts(x, {'end_pos': [float(x[dn]) for dn in end_dim_names]}))

    '''
    entry_ranges = entries.map(lambda x: ([x[importance_field]] + x['pos'] + x['end_pos'],
                                          [x[importance_field]] + x['pos'] + x['end_pos']))


    reduced_entry_ranges = entry_ranges.reduce(reduce_range)

    #mins = reduced_entry_ranges[0][1:1+len(dim_names)]
    #maxs = reduced_entry_ranges[1][1+len(dim_names):]
    '''

    max_width = max(map(lambda x: x[1] - x[0], zip(mins, maxs)))

    zoom_level = 0
    tile_entries = sc.parallelize([])

    if max_zoom is None:
        # hopefully it'll end up being smaller than that
        max_zoom = 1024

    if importance_field is None:
        importance_field = dim_names[0]

    # add zoom levels until we either reach the maximum zoom level or
    # have no tiles that have more entries than max_entries_per_tile
    while zoom_level <= max_zoom:
        tile_width = max_width / 2**zoom_level

        if tile_width == 0:
            tile_width = 1

        def place_in_tile(entry):
            tile_positions = []

            for (i,mind) in enumerate(mins):
                curr_pos = entry['pos'][i]
                dimension_tile_positions = [int((curr_pos - mind) / tile_width)]

                # if this feature spans multiple tiles, its end will be after the start
                # of the next tile
                next_tile_start_pos = mind + ((curr_pos - mind) // tile_width + 1) * tile_width

                while next_tile_start_pos < entry['end_pos'][i]:
                    # spans into the next tile
                    dimension_tile_positions += [int((next_tile_start_pos - mind) // tile_width)]
                    next_tile_start_pos += tile_width

                tile_positions += [dimension_tile_positions]

            # transpose the tile positions
            output_positions = [(tuple([zoom_level] + p), [entry]) for p in map(list, zip(*tile_positions))]

            return output_positions

        current_tile_entries = entries.flatMap(place_in_tile)
        current_max_entries_per_tile = max(current_tile_entries.countByKey().values())
        tile_entries = tile_entries.union(current_tile_entries)

        if adapt_zoom and current_max_entries_per_tile <= max_entries_per_tile:
            max_zoom = zoom_level
            break

        zoom_level += 1

    # all entries are broked up into ((tile_pos), [entry]) tuples
    # we just need to reduce the tiles so that no tile contains more than
    # max_entries_per_tile entries
    # (notice that [entry] is an array), this format will be important when
    # reducing to the most important values
    def reduce_values_by_importance(entry1, entry2):
        if reverse_importance:
            combined_entries = sorted(entry1 + entry2,
                    key=lambda x: -float(x[importance_field]))
        else:
            combined_entries = sorted(entry1 + entry2,
                    key=lambda x: float(x[importance_field]))
        return combined_entries[:max_entries_per_tile]

    reduced_tiles = tile_entries.reduceByKey(reduce_values_by_importance)

    tileset_info = {}
    tileset_info['max_importance'] = entries.map(lambda x: float(x[importance_field])).reduce(reduce_max)
    tileset_info['min_importance'] = entries.map(lambda x: float(x[importance_field])).reduce(reduce_min)

    tileset_info['min_pos'] = mins
    tileset_info['max_pos'] = maxs

    tileset_info['max_zoom'] = max_zoom
    tileset_info['max_width'] = max_width

    return {"tileset_info": tileset_info, "tiles": reduced_tiles}

def reduce_max(a,b):
    return max(a,b)

def reduce_min(a,b):
    return min(a,b)

def reduce_range(range_a, range_b):
    '''
    Get the range of two ranges, a and b.

    Example: a = [(1,3),(2,4)]
             b = [(0,5),(7,10)]
             ------------------
          result [(0,3),(7,10)]
    '''
    (mins_a, maxs_a) = range_a
    (mins_b, maxs_b) = range_b
    mins_c = [min(a,b) for (a,b) in zip(mins_a, mins_b)]
    maxs_c = [max(a,b) for (a,b) in zip(maxs_a, maxs_b)]

    return mins_c, maxs_c

def reduce_bins(bins_a, bins_b):
    for bin_pos in bins_b:
        bins_a[bin_pos] += bins_b[bin_pos]
    return bins_a

def reduce_sum(a,b):
    return a + b

def make_tiles_by_binning(sc, entries, dim_names, max_zoom, value_field='count', 
        importance_field='count', resolution=None,
        bins_per_dimension=1,
        num_histogram_bins=1000):
    '''
    Create tiles by calculating tile indeces.

    The first tile will encompass the entire data set, and each
    subsequent tile will contain a fraction of that data.

    :param dim_names: The names of the fields containing the positions of the data
    :param max_zoom: The maximum zoom level
    :param resolution: The resolution of the data
    :param aggregate_tile: Condense the entries in a given tile 
        (should operate on a single tile)
    '''
    max_data_in_sparse = bins_per_dimension ** len(dim_names) / 30.

    epsilon = 0.0000    # for calculating the max width so that all entries
                        # end up in the same top_level bucket

    # if the resolution is provided, we could go from the bottom up
    # and create zoom_widths that are multiples of the resolutions
    def consolidate_positions(entry):
        '''
        Place all of the dimensions in one array for this entry.
        '''
        new_entry = { 'pos': [float(entry[dn]) for dn in dim_names],
                      'value': float(entry[value_field]),
                      'importance': float(entry[importance_field]) }
        return new_entry

    def add_sparse_tile_metadata(tile_stuff):
        (tile_id, tile_entries_iterator) = tile_stuff
        z = tile_id[0]
        tile_width = max_width / 2 ** z
        bin_width = tile_width / bins_per_dimension

        # calculate where the tile values start along each dimension
        '''
        tile_start_pos = map(lambda (z,(m,x)): m + x * tile_width, 
                it.izip(it.cycle([tile_id[0]]), zip(mins, tile_id[1:])))

        tile_end_pos = map(lambda (z,(m,x)): m + (x+1) * tile_width, 
                it.izip(it.cycle([tile_id[0]]), zip(mins, tile_id[1:])))
        '''

        shown = []
        for (bin_pos, bin_val) in tile_entries_iterator:
            #pos = map(lambda(md, x): md + x * bin_width, zip(tile_start_pos, bin_pos))
            shown += [{'pos': bin_pos, 'value' : bin_val}]

        # caclulate where the tile values end along each dimension
        '''
        tile_data = {'shown': shown,
                     'zoom': tile_id[0],
                     'tile_start_pos': tile_start_pos,
                     'tile_end_pos': tile_end_pos}
        '''
        tile_data = shown

        return (tile_id, tile_data)

    # add the tile meta-data
    def add_dense_tile_metadata(tile_stuff):
        '''
        Add the tile's start and end data positions.
        '''
        (tile_id, tile_entries_iterator) = tile_stuff
        shown = []

        initial_values = [0.0] * (bins_per_dimension ** len(dim_names))

        for (bin_pos, val) in tile_entries_iterator:
            index = sum([bp * bins_per_dimension ** i for i,bp in enumerate(bin_pos)])
            initial_values[index] = val

        shown = initial_values

        # caclulate where the tile values end along each dimension
        tile_data = shown

        return (tile_id, tile_data)

    def add_tile_metadata(tile):
        if len(tile[1]) > max_data_in_sparse:
            return add_dense_tile_metadata(tile)
        else:
            return add_sparse_tile_metadata(tile)

    def dense_range(tile_value):
        return (min(tile_value), max(tile_value))

    def sparse_range(tile_value):
        values = [x['value'] for x in tile_value]

        return (0, max(values))

    def tile_pos_to_string(tile):
        (key, tile_value) = tile
        if len(tile_value) > max_data_in_sparse:
            (min_value, max_value) = dense_range(tile_value)
            output_str = {'dense': tile_value}
        else:
            (min_value, max_value) = sparse_range(tile_value)
            output_str = {'sparse': tile_value}

        output_str['min_value'] = min_value
        output_str['max_value'] = max_value

        return (key, output_str)

    entries = entries.map(consolidate_positions)
    tiled_entries = entries.map(lambda x: (0, x))
    all_tiles = sc.parallelize([])

    tileset_info = {}

    entry_ranges = entries.map(lambda x: ([x['value'], x['importance']] + x['pos'],
                                         ([x['value'], x['importance']] + x['pos'])))
    reduced_entry_ranges = entry_ranges.reduce(reduce_range)

    tileset_info['max_value'] = reduced_entry_ranges[1][0]
    tileset_info['min_value'] = reduced_entry_ranges[0][0]

    tileset_info['max_importance'] = reduced_entry_ranges[1][1]
    tileset_info['min_importance'] = reduced_entry_ranges[0][1]

    mins = reduced_entry_ranges[0][2:]
    maxs = reduced_entry_ranges[1][2:]

    value_histogram = []
    bin_size = (tileset_info['max_value'] - tileset_info['min_value']) / num_histogram_bins

    if bin_size == 0:
        bin_size = 1   # min_value == max_value

    histogram_counts = entries.map(lambda x: (int((x['value'] - tileset_info['min_value']) / bin_size), 1)).countByKey().items()
    histogram = {"min_value": tileset_info['min_value'],
                 "max_value": tileset_info['max_value'],
                 "counts": histogram_counts}

    # O(n) get the maximum and minimum bounds of the data set
    #(mins, maxs) = data_bounds(entries, len(dim_names))
    max_width = max(map(lambda x: x[1] - x[0] + epsilon, zip(mins, maxs)))

    if resolution is not None:
        # r * 2 ** n-1 < max_width < r * 2 ** n
        # we need a max width that is a multiple of the resolution, the bin size
        # and a power of 2. 
        if bins_per_dimension is None:
            bins_per_dimension = 1

        bins_to_display_at_max_resolution = max_width / resolution / bins_per_dimension
        max_max_zoom = math.ceil(math.log(bins_to_display_at_max_resolution) / math.log(2.))

        max_width = resolution * bins_per_dimension * 2 ** max_max_zoom

        if max_max_zoom < 0:
            max_max_zoom = 0

        if max_max_zoom < max_zoom:
            max_zoom = int(max_max_zoom)

    # get all the different zoom levels
    zoom_levels = range(max_zoom+1)

    zoom_width = max_width / 2 ** max_zoom
    bin_width = zoom_width / bins_per_dimension

    tileset_info['min_pos'] = mins
    tileset_info['max_pos'] = maxs

    tileset_info['max_zoom'] = max_zoom
    tileset_info['max_width'] = max_width

    tileset_info['data_granularity'] = resolution
    tileset_info['bins_per_dimension'] = bins_per_dimension

    def place_positions_at_origin(entry):
        #new_pos = map(lambda (x, mx): x - mx, zip(entry['pos'], mins))
        new_pos = [x - mx for (x,mx) in zip(entry['pos'], mins)]
        return (new_pos, entry['value'])

    # add a bogus tile_id for downstream processing
    bin_entries = entries.map(place_positions_at_origin)
    total_bins = 0
    total_tiles = 0
    total_start_time = time.time()

    for zoom_level in range(0, max_zoom+1)[::-1]:
        tile_width = max_width / 2 ** zoom_level
        bin_width = tile_width / bins_per_dimension
        start_time = time.time()

        def place_in_bin(pos_value):
            (prev_pos, value) = pos_value
            new_bin_pos = tuple([int(int(x / bin_width) * bin_width) for x in prev_pos])

            return (new_bin_pos, value)

        def place_in_tile(bin_pos_value):
            (bin_pos, value) = bin_pos_value

            # we have a bin position and we need to place it in a tile
            tile_pos = [int(x / tile_width) for x in bin_pos]
            tile_mins = [x * tile_width for x in tile_pos]
            bin_in_tile = [int((bin_pos[i] - mind) / bin_width) for (i,mind) in enumerate(tile_mins)]

            return (tuple([zoom_level] + tile_pos), [(bin_in_tile, value)])


        bin_entries = bin_entries.map(place_in_bin).reduceByKey(reduce_sum)
        '''
        bin_count = bin_entries.count()
        total_bins += bin_count
        '''

        tile_entries = bin_entries.map(place_in_tile)
        tile_entries = tile_entries.reduceByKey(reduce_sum)
        '''
        tile_count = tile_entries.count()
        total_tiles += tile_count
        '''

        '''

        def place_in_tile(entry):
            tile_pos = tuple( [zoom_level] + 
                    map(lambda (i, mind): int((entry['pos'][i] - mind) / tile_width),
                           enumerate(mins)))

            tile_mins = map(lambda (z,(m,x)): m + x * (max_width / 2 ** z), 
                    it.izip(it.cycle([tile_pos[0]]), zip(mins, tile_pos[1:])))

            bin_pos = map(lambda (i, mind): int((entry['pos'][i] - mind) / bin_width),
                          enumerate(tile_mins))
            bin_dict = col.defaultdict(float)
            bin_dict[tuple(bin_pos)] = entry[value_field]
            return ((tile_pos), bin_dict)
        
        tile_entries = entries.map(place_in_tile).reduceByKey(reduce_bins)
        '''
        tiles_with_meta = tile_entries.map(add_tile_metadata)
        tiles_with_meta_string = tiles_with_meta.map(tile_pos_to_string)
        all_tiles = all_tiles.union(tiles_with_meta_string)



        end_time = time.time()


    total_end_time = time.time()
    #total_entries = entries.count()

    return {"tileset_info": tileset_info, "tiles": all_tiles, "histogram": histogram}



#!/usr/bin/python

import argparse
import clodius.save_tiles as cst
import functools as ft
import itertools as it
import collections as col
import math
import negspy.coordinates as nc
import numpy as np
import os
import os.path as op
import sortedcontainers as sco
import sys
import time

import multiprocessing as mpr
import traceback

sys.excepthook = cst.handle_exception

def default_tile(length):
    return np.zeros(length)

def create_tiles(q, first_lines, input_source, position_cols, value_pos, max_zoom, 
        bins_per_dimension, tile_saver, expand_range, ignore_0, tileset_info, max_width,
        triangular=False, max_queue_size = 40000, print_status=False):
    active_tiles = col.defaultdict(sco.SortedList)
    max_data_in_sparse = bins_per_dimension ** len(position_cols) // 5.
    tile_contents = col.defaultdict(lambda: col.defaultdict(lambda: col.defaultdict(ft.partial(default_tile, length=len(value_pos)))))
    smallest_width = max_width // (2 ** max_zoom)

    prev_line_num = None
    prev_time = time.time()
    start_time = time.time()

    calculated_tile_bin_positions = col.defaultdict(dict)
    for x in it.product(range(2), repeat=len(position_cols)):
        for y in it.product(range(bins_per_dimension), repeat = len(position_cols)):
            calculated_tile_bin_positions[x][y] = tuple([ int((bins_per_dimension * tp + bp) // 2) % bins_per_dimension for  (tp, bp) in zip(x, y)])

    def add_to_next_tile(zoom_level, tile_position, tile_bins):
        next_tile_position = tuple([tp // 2 for tp in tile_position])
        new_tile_contents = tile_contents[zoom_level-1][next_tile_position]

        if next_tile_position not in active_tiles[zoom_level - 1]:
            active_tiles[zoom_level - 1].add(next_tile_position)

        tile_mod_pos = tuple([tp % 2 for tp in tile_position])
        calc_bin_poss = calculated_tile_bin_positions[tile_mod_pos]

        for bin_position,bin_value in tile_bins.items():
            #old_abs_pos = [tp * bins_per_dimension + bp for (tp, bp) in zip(tile_position, bin_position)]
            #new_bin_pos = tuple([int(op / 2) % bins_per_dimension for op in old_abs_pos])
            #print "tile_position:", tile_position, bin_position
            '''
            for i in range(len(tile_position)):
                tp = tile_position[i]
                bp = bin_position[i]
                new_bin_pos_list[i] = int((bins_per_dimension * tp + bp) / 2) % bins_per_dimension
            new_bin_pos = tuple(new_bin_pos_list)

            new_bin_pos = calc_bin_poss[bin_position]
            print "tp, bp:", tile_position, bin_position, new_bin_pos, calc_bin_pos
            if calc_bin_pos != new_bin_pos:
                print "Wrong calc_bin_pos!"
                sys.exit(1)
            #new_bin_pos = (0,)
            if zoom_level < 3:
                print >>sys.stderr, "adding to next tile:", zoom_level-1, next_tile_position, "new_bin_pos:", new_bin_pos
            '''
            try:
                new_tile_contents[calc_bin_poss[bin_position]] += bin_value
                #print("ntc:", new_tile_contents[calc_bin_poss[bin_position]])
            except KeyError:
                print("calc_bin_poss.keys():", calc_bin_poss.keys())
                print("tile_mod_pos:", tile_mod_pos, tile_position)
                raise

    for line_num,line in enumerate(it.chain(first_lines, input_source)):
        # see if we have to expand any coordinate ranges
        # bedFiles are examples
        # chr1 100 102 0.5
        # means that we need to expand that to
        # chr1 100 101 0.5
        # chr1 101 102 0.5
        all_line_parts = []

        all_line_parts = [[p for p in line.strip().split()]]

        for line_parts in all_line_parts:
            if ignore_0:
                if line_parts[value_pos][0] == "0":
                    continue
            value = [float(line_parts[vp]) for vp in value_pos]
            #print "entry_pos:", entry_pos, value

            if print_status is not None:
                if line_num != prev_line_num and line_num % print_status == 0:
                    time_str = time.strftime("%Y-%m-%d %H:%M:%S")
                    print( "current_time:", time_str, "line_num:", line_num, "time:", int(1000 * (time.time() - prev_time)), "total_time", int(time.time() - start_time))

                prev_time = time.time()
            prev_line_num = line_num

            if triangular:
                entry_pos = sorted([float(line_parts[p-1]) for p in position_cols])
            else:
                entry_pos = [float(line_parts[p-1]) for p in position_cols]

            tileset_info['max_value'] = [max(x,y) for x,y in zip(tileset_info['max_value'], value)]
            tileset_info['min_value'] = [min(x,y) for x,y in zip(tileset_info['min_value'], value)]

            entry_poss = [entry_pos]
            if expand_range is not None:
                end_pos = int(line_parts[expand_range[1]-1])
                #print("ep", int(entry_pos[0]), end_pos -  int(entry_pos[0]) + 1)
                for i in range(int(entry_pos[0]), end_pos):
                    new_entry = entry_pos[::]
                    new_entry[0] = i
                    entry_poss += [new_entry]

            # the bin within the tile as well as the tile position
            # place this data point in the highest resolution tile that we can
            for entry_pos in entry_poss:
                current_bin = tuple([int(ep // (smallest_width // bins_per_dimension)) % bins_per_dimension for ep in entry_pos])
                current_tile = tuple([int(ep // smallest_width) for ep in entry_pos])

                # we haven't seen this tile before so we save it, in case more data points need to be placed in it
                if current_tile not in active_tiles[max_zoom]:
                    active_tiles[max_zoom].add(current_tile)

                # place the data in the tile
                tile_contents[max_zoom][current_tile][current_bin] += value
            
            # go through all the tiles and check which ones we're done with
            # start with the max zoom
            for zoom_level in range(0, max_zoom+1)[::-1]:
                #print "zoom_level:", zoom_level, "active_tiles length:", len(active_tiles[zoom_level]), "total_time", int(time.time() - start_time)
                current_tile = tuple([int(ep // (smallest_width * 2 ** (max_zoom - zoom_level)) ) for ep in entry_pos])

                # keep track of whether any tiles were finished at the previous level
                # we can't complete a lower zoom tile unless we complete a higher zoom one
                finished_tile = False
                while len(active_tiles[zoom_level]) > 0:
                    '''
                    if zoom_level < max_zoom:
                        print "at:", zoom_level, "len(at):", len(active_tiles[zoom_level]), active_tiles[zoom_level][0], current_tile
                    '''
                    if active_tiles[zoom_level][0][0] < current_tile[0]:
                        tile_position = active_tiles[zoom_level][0]
                        tile_value = tile_contents[zoom_level][tile_position]
                        tile_bins = tile_contents[zoom_level][tile_position]

                        #print "tile_bins:", zoom_level, tile_position, tile_bins

                        # make sure old requests get saved before we create new ones
                        already_sleeping = False
                        while q.full(): #q.qsize() > max_queue_size:
                            if print_status:
                                if already_sleeping:
                                    sys.stdout.write('.')
                                    sys.stdout.flush()
                                else:
                                    sys.stdout.write('sleeping.')
                                    sys.stdout.flush()
                                    already_sleeping = True
                            time.sleep(0.5)
                        if already_sleeping:
                            if print_status:
                                sys.stdout.write('.\n')

                        q.put((zoom_level, active_tiles[zoom_level][0], tile_bins))
                        '''
                        if zoom_level < max_zoom:
                            print "putting tile"
                        '''
                        finished_tile = True

                        #print "zoom_level:", zoom_level, max_zoom
                        if zoom_level > 0:
                            add_to_next_tile(zoom_level, tile_position, tile_bins)

                        # create a lower zoom level tile

                        del tile_contents[zoom_level][tile_position]
                        del active_tiles[zoom_level][0]
                    else:
                        break

                if not finished_tile:
                    break

            '''
            for zoom_level, tile_width in zip(range(0, max_zoom+1), tile_widths):


                # which bins will never be touched again?
                # all bins at the current zoom level where (entry_pos[0] / tile_width) < current_bin[0]
                while len(active_tiles[zoom_level]) > 0:
                    if active_tiles[zoom_level][0][0] < current_tile[0]:
                        tile_position = active_tiles[zoom_level][0]
                        tile_value = tile_contents[zoom_level][tile_position]
                        tile_bins = tile_contents[zoom_level][tile_position]

                        while q.qsize() > 40000:
                            time.sleep(1)

                        q.put((zoom_level, active_tiles[zoom_level][0], tile_bins))

                        del tile_contents[zoom_level][tile_position]
                        del active_tiles[zoom_level][0]
                    else:
                        break
            '''

    for zoom_level in range(0, max_zoom+1)[::-1]:
        for tile_position in active_tiles[zoom_level]:
            tile_value = tile_contents[zoom_level][tile_position]
            tile_bins = tile_contents[zoom_level][tile_position]

            q.put((zoom_level, tile_position, tile_bins))
            
            if zoom_level > 0:
                add_to_next_tile(zoom_level, tile_position, tile_bins)

    tile_saver.save_tile({'tile_id': 'tileset_info', 
                          'tile_value': tileset_info})


    while not q.empty():
        #print "qsize:", q.qsize()
        time.sleep(1)

    tile_saver.flush()
    return tileset_info

def main():
    usage = """
    python make_tiles.py input_file

    Create tiles for all of the entries in the JSON file.
    """
    parser = argparse.ArgumentParser()

    #parser.add_argument('-o', '--options', dest='some_option', default='yo', help="Place holder for a real option", type='str')
    #parser.add_argument('-u', '--useless', dest='uselesss', default=False, action='store_true', help='Another useless option')
    parser.add_argument('--min-pos',
                        help="The minimum range for the tiling")
    parser.add_argument('--max-pos',
                        help="The maximum range for the tiling")
    parser.add_argument('--assembly', default=None)
    parser.add_argument('-r', '--resolution', help="The resolution of the data", 
                        default=None, type=int )
    parser.add_argument('-k', '--position-cols', help="The position columns (defaults to all but the last, 1-based)", default=None)
    parser.add_argument('-v', '--value-pos', help='The value column (defaults to the last one, 1-based)', default=None, type=str)
    parser.add_argument('-z', '--max-zoom', help='The maximum zoom value', default=None, type=int)
    parser.add_argument('--expand-range', help='Expand ranges of values')
    parser.add_argument('--ignore-0', help='Ignore ranges with a zero value', default=False, action='store_true')
    parser.add_argument('-b', '--bins-per-dimension', default=1,
                        help="The number of bins to consider in each dimension",
                        type=int)
    parser.add_argument('-e', '--elasticsearch-url', default=None,
                        help="The url of the elasticsearch database where to save the tiles")
    parser.add_argument('-f', '--columnfile-path', default=None,
                        help="The path to the column file where to save the tiles")
    parser.add_argument('-n', '--num-threads', default=4, type=int)
    parser.add_argument('--triangular', default=False, action='store_true')
    parser.add_argument('--log-file', default=None)
    parser.add_argument('--max-queue-size', default=40000, type=int)
    parser.add_argument('--print-status', default=None, type=int)

    args = parser.parse_args()

    if args.resolution is None and args.max_zoom is None:
        print >>sys.stderr, "One of --resolution and --max-zoom must be set"
        sys.exit(1)

    first_line = sys.stdin.readline()
    first_line_parts = first_line.strip().split()
    if len(first_line_parts) == 0:
        print("ERROR: no input")
        return

    if args.position_cols is not None:
        position_cols = list(map(int, args.position_cols.split(',')))
    else:
        position_cols = None

    # if specific position columns aren't specified, use all but the last column
    if position_cols is None:
        position_cols = list(range(1,len(first_line_parts)))

    if args.assembly is not None:
        mins = [1 for p in position_cols]
        maxs = [nc.get_chrominfo(args.assembly).total_length for p in position_cols]
    else: 
        mins = [float(p) for p in args.min_pos.split(',')]
        maxs = [float(p) for p in args.max_pos.split(',')]

    max_width = max([b - a for (a,b) in zip(mins, maxs)])

    if args.expand_range is not None:
        expand_range = list(map(int, args.expand_range.split(',')))
    else:
        expand_range = None


    if args.max_zoom is None:
        # determine the maximum zoom level based on the domain of the data
        # and the resolution
        bins_to_display_at_max_resolution = max_width // args.resolution // args.bins_per_dimension
        max_max_zoom = math.ceil(math.log(bins_to_display_at_max_resolution) / math.log(2.))


        if max_max_zoom < 0:
            max_max_zoom = 0

        max_zoom = int(max_max_zoom)
    else:
        max_zoom = args.max_zoom

    #print("max_zoom:", max_zoom)
    max_width = args.resolution * args.bins_per_dimension * 2 ** max_zoom
    smallest_width = args.resolution * args.bins_per_dimension

    value_pos = args.value_pos
    

    # if there's not column designated as the value column, use the last column
    if value_pos is None:
        value_pos = [len(first_line_parts)-1]
    else:
        value_pos = [int(vp) - 1 for vp in value_pos.split(',')]

    max_data_in_sparse = args.bins_per_dimension ** len(position_cols) // 10

    '''
    if args.elasticsearch_url is not None:
        tile_saver = cst.ElasticSearchTileSaver(max_data_in_sparse,
                                                args.bins_per_dimension,
                                                num_dimensions = len(position_cols),
                                                es_path = args.elasticsearch_url)
    else:
        tile_saver = cst.EmptyTileSaver(max_data_in_sparse,
                                        args.bins_per_dimension,
                                        num_dimensions = len(position_cols))
    '''

    print("maxs:", maxs, "max_zoom:", max_zoom, "max_data_in_sparse:", max_data_in_sparse, "url:", args.elasticsearch_url)

    #bin_counts = col.defaultdict(col.defaultdict(int))
    q = mpr.Queue(maxsize=args.max_queue_size)

    tilesaver_processes = []
    finished = mpr.Value('b', False)
    if args.elasticsearch_url is not None:    
        tile_saver = cst.ElasticSearchTileSaver(max_data_in_sparse,
                                                args.bins_per_dimension,
                                                len(position_cols),
                                                args.elasticsearch_url,
                                                args.log_file,
                                                args.print_status,
                                                initial_value = [0. for vp in value_pos])
    else:
        tile_saver = cst.ColumnFileTileSaver(max_data_in_sparse,
                                                args.bins_per_dimension,
                                                len(position_cols),
                                                args.columnfile_path,
                                                args.log_file,
                                                args.print_status,
                                                initial_value = [0. for vp in value_pos])

    for i in range(args.num_threads):
        p = mpr.Process(target=cst.tile_saver_worker, args=(q, tile_saver, finished))

        p.daemon = True
        p.start()
        tilesaver_processes += [(tile_saver, p)]

    tileset_info = {'max_value': [0 for vp in value_pos],
                    'min_value': [0 for vp in value_pos],
                    'min_pos': mins,
                    'max_pos': maxs,
                    'max_zoom': max_zoom,
                    'bins_per_dimension': args.bins_per_dimension,
                    'max_width': max_width}


    tile_saver.save_tile({'tile_id': 'tileset_info', 
                          'tile_value': tileset_info})
    tile_saver.flush()

    try:
        tileset_info = create_tiles(q, [first_line], sys.stdin, position_cols, value_pos, 
                max_zoom, args.bins_per_dimension, tile_saver, expand_range,
                args.ignore_0, tileset_info, max_width, args.triangular, args.max_queue_size,
                print_status=args.print_status)
    except KeyboardInterrupt:
        for (ts, p) in tilesaver_processes:
            ts.flush()
            p.terminate()
            p.join()
        raise

    finished.value = True
    # wait for the worker processes to finish
    for (ts, p) in tilesaver_processes:
        p.join()

    print("tileset_info:", tileset_info)
    tile_saver.save_tile({'tile_id': 'tileset_info', 
                          'tile_value': tileset_info})
    tile_saver.flush()
         

if __name__ == '__main__':
    main()





import sqlite3

def get_tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    c = conn.cursor()

    row = c.execute("SELECT * from tileset_info").fetchone();
    tileset_info = {
            'zoom_step': row[0],
            'max_length': row[1],
            'assembly': row[2],
            'chrom_names': row[3],
            'chrom_sizes': row[4],
            'tile_size': row[5],
            'max_zoom': row[6],
            'max_width': row[7],
            "min_pos": [1],
            "max_pos": [row[1]]
            }
    conn.close()

    return tileset_info

def get_tile(db_file, zoom, tile_x_pos):
    tileset_info = get_tileset_info(db_file)

    conn = sqlite3.connect(db_file)
    c = conn.cursor()
    tile_width = tileset_info['max_width'] / 2 ** zoom

    tile_start_pos = tile_width * tile_x_pos
    tile_end_pos = tile_start_pos + tile_width

    print("tile_start_pos:", tile_start_pos, 'tile_end_pos:', tile_end_pos)

    query = '''
    SELECT chrOffset, fields from intervals,position_index 
    where 
    intervals.id=position_index.id and zoomLevel <= {} and rEndPos >= {} and rStartPos <= {}
    '''.format(zoom, tile_start_pos, tile_end_pos)

    print("query", query)

    rows = c.execute(query).fetchall()
    print("rows:", rows)


    # add the position offset to the returned values
    rows = [r[1].split('\t') + [r[0]] for r in rows]
    conn.close()

    return rows



from distutils.core import setup
from distutils.extension import Extension

from Cython.Build import cythonize
from Cython.Distutils import build_ext

import numpy 

setup(name='clodius',
     include_dirs=[numpy.get_include()],
      ext_modules = cythonize([Extension("clodius.fast", ["clodius/fast.pyx"])]),
      version='0.3.1',
      description='Tile generation of big data',
      author='Peter Kerpedjiev',
      author_email='pkerpedjiev@gmail.com',
      url='',
      packages=['clodius'],
      setup_requires = [
          'cython>=0.x'
          ],
      scripts=['scripts/make_single_threaded_tiles.py', 'scripts/tile_bigWig.py']
     )
